/**
 * 
 * 27C160 (and probably 27C400 and 800) EPROM reader/writer. Based on meeprommer.
 * 
 * This is a hack of meeprommer-am29f040 so that you can program a 27C160
 *
 * - The wiring is much the same as meeprommer, but we use a 3rd 74HC595
 * - You need two extra power supplies and will have to manually hook them in at
 *   the appropriate moments
 *    - You need one PSU @6.25V, and one @12.5V. The GND outputs of these 
 *      should to tied to the GND of the Arduino and EPROM and HC595s. But 
 *      the +ve output voltages should be initially disconnected.
 *      
 *   I used a cheap adjustable 'DC Boost' PSU adapter to get 12.5V from 5V inputs, and a LM2596 to step   
 *   down from 12V to 6.25V. Make sure these are adjusted correctly before you apply
 *   them to the EPROM.
 * - You initially wire the EPROM with Vcc @5V (powered from the arduino), and BYTE/Vpp tied to GND
 * - You should be able to read the content of the EPROM this way
 * - When you want to program the EPROM you need to carefully do this
 *     - quickly change the Vcc pin of the EPROM so that it is running off the 6.25V +ve supply
 *     - quickly switch the BYTE/Vpp pin from GND to +12.5V
 * - Now run the xmodem transfer command to write a file to the EPROM.
 * - When the transfer finishes;
 *     - quickly switch the BYTE/Vpp pin from 12.5V to GND
 *     - quickly swich the Vcc pin from 6.25V to 5V
 * - Try using the r command to read back what you just wrote. Try running the m command to md5 
 *   what you just wrote.
 *   
 * WARNING: Don't put an original Amiga kickstart mask ROM in as a test to see if you can read it.
 *          While the 27C400/800/160 EPROMS have a special BYTE read mode, the mask ROMs may not 
 *          support it, and you may do damage to the mask ROM as a result
 *
 * kernelcrash.com
 *
 * YOU MUST CHANGE THE BOARD CONFIG TO DO A 256 BYTE SERIAL RECEIVE BUFFER!!!!!
 * http://www.hobbytronics.co.uk/arduino-serial-buffer-size
 * 
 * 
 * HISTORY
 * =======
 * 
 *  - based on meeprommer from https://github.com/mkeller0815/MEEPROMMER
 *  - based on meeprommer-am29f040 v0.02
 *
 * v0.01 - removed the 74LS374 alternative code
                       
 * Basic wiring
 *
 * - Use two 74HC595s to generate the address bus as per the original meeprommer, with one addition
 *   being the Q7 output of the high address bus 74HC595 goes to A15 of the EPROM. The special Q7' (p9)
 *   output of the low address latch is connected to DS of the high address latch
 * - Add a 3rd 74HC595 chained off the high address 74HC595 (ie. Q7' of the high address latch goes
 *   to DS of this 3rd 74HC595). The Q7 to Q0 outputs of this 74HC595 
 *   go to the D15 to D8 pins of the 27C160 EPROM (sometimes labelled as Q15 to Q8)
 * 
 * 
 * 74HC595 version
 * ===============
 * 
 * D2 to D9 of the arduino uno are wired to the databus D0 to D7 of the 27C160
 * 
 * A0 is wired as per meeprommer to DS pin of the low (A7 to A0) 74HC595 
 *     (I will call this the low address 74HC595)
 * A1 is wired as per meeprommer to the ST_CP pin of all three 74HC595s
 * A2 is wired as per meeprommer to the SH_CP pin of all three 74HC595s
 * 
 * A3 of arduino goes to E* (CE*) on the 27C160
 * A4 of arduino goes to G* (OE*) on the 27C160
 * A5 of arduino goes to OE* of the new 3rd 74HC595
 * 
 * The BYTE/Vpp pin is manually switched. On power on tie it LOW for BYTE mode
 * 
 * D10 of arduino goes to A16 of 27C160
 * D11 of arduino goes to A17 of 27C160
 * D12 of arduino goes to A18 of 27C160
 * There is no arduino pin going to A19 of the 27C160. Just tie A19 high or low to program the upper or lower 1MByte
 * 
 * The Q0 to Q7 outputs of the low address 74HC595 go to 
 *     A0 to A7 of the 27C160 ROM
 *     
 * The Q0 to Q7 ouptuts of the high address 74HC595 (ie. the other one) 
 *     go to A8 to A15 of the 27C160 ROM (in the meeprommer schematic Q7 
 *     is not connected due to the smaller size of the EEPROM.
 *
 * The Q0 to Q7 outputs of the new 3rd 74HC595 go to the Q8 to Q15 pins of the 27C160
 *     
 * refer to the meeprommer schematic for how everything else is wired up. 
 * 
 * Commands   
 *    a nnnnn   - sets the  address bus to hex nnnnn
 *    w nnnnn   - write data to the EPROM at hex address nnnnn using an xmodem-crc transfer
 *    r nnnnn mmmmm - read from EPROM at address nnnnn for mmmmm bytes (all in hex)
 *    t nnnnn 0xxyy - write two bytes (xx and yy) to EPROM at nnnnn (for testing)
 *    m nnnnn mmmmm - do an md5 hash of the content of the 
 *         EPROM starting at nnnnn and being mmmmm bytes long
 *    b nnnnn mmmmm - do a 'blank check' starting at nnnnn and going for mmmmm bytes. 
 *         It looks for FF bytes, and shows a message on the first one that is not FF
 *    
 * ENDIANESS/BYTE SWAPPING
 * =======================
 *
 * Since we are programming a 16 bit wide EPROM we need to care about endianness.
 * In the config.h file you can set
 * 
 *  #define LITTLEENDIAN false
 *
 *  OR
 *
 *  #define LITTLEENDIAN true
 *
 *  - if 'true' then the 1st byte in your xmodem stream goes to D7 to D0 and the 2nd byte to D15-D8. ie. you are
 *    writing for a little endian system
 *  - if 'false' then the 1st byte in your xmodem stream goes to D15-D8, and the 2nd byte to D7-D0 and so on. ie. you
 *    are writing for a big endian system. An Amiga is big endian.
 *
 * For a big endian system, if you read back what you just wrote it will look byte-swapped. Doing an md5 will also
 * give a different result. If you want to check your md5, then you would need to byte swap the file you just wrote
 * to the EPROM. For example;
 *
 *    dd conv=swab if=kick.rom of=kick-byteswapped.rom
 *    md5sum kick-byteswapped.rom
 *

 * INSTRUCTIONS
 *
 * Make a decision about the LITTLEENDIAN define and compile and upload to the Uno
 * Wire up arduino uno as per notes above. 
 * Make sure the EPROM Vcc is at 5V, and BYTE/Vpp is tied LOW
 * Make sure you've made the necessary changes to support 256byte serial buffers.
 * YEAH YOU HAVE TO DO THAT.
 * Load the sketch in the arduino IDE. Edit the config.h
 * Program the sketch using the arduino IDE. Turn off the arduino IDE and start
 * a serial terminal program with 115200 bps and 8N1. Hopefully hit ENTER and you should
 * get a '%' prompt.
 *
 * You can read your 27C160 with
 * 
 *  % r 00000 00080
 *
 * (NOTE: the 00080 number of bytes bit is optional)
 * 
 *  
 * But the main command is to do an xmodem transfer from your PC to write to the EPROM.
 * Lets say you have a binary file on your PC that you want to write to the start of 
 * your EPROM. Do this
 * 
 *  % w 00000
 * 
 * You get some blurb about changing Vcc and the Vpp pin. First quickly change Vcc from 5V to 6.25V
 * Then change the Vpp pin to 12.5V.
 * Then choose the option in your terminal program to start an xmodem transfer. In minicom
 * this is 'ctrl-a s xmodem'. And then choose the file to send and let it transfer. When its
 * done. Hit ENTER a few times to get your prompt back and perhaps use the 'r' command' to 
 * check that something was written.
 * 
 *  % m 00000 08000
 *  
 *  md5sum from 00000 for 08000 bytes (ie. first 32K of the EPROM). Can take a while. If you are in 
 * big endian mode then the md5 will never match your source file. YOu need to make a byte swapped version
 * of your source file and md5 it.
 *  
 *   dd conv=swab if=testkick31.rom of=testkick31-swapped.rom
 *  
 *
 * Orig meeprommer author details
 *
 * $Author: mario $
 * $Date: 2013/05/05 11:01:54 $
 * $Revision: 1.2 $
 *
 * This software is freeware and can be modified, reused or thrown away without any restrictions.
 *
 * Use this code at your own risk. I'm not responsible for any bad effect or damages caused by this software!!!
 *
 **/

#include "config.h"
#include "md5stuff.h"

//#define VERSIONSTRING "MEEPROMMER $Revision: 1.2 $ $Date: 2013/05/05 11:01:54 $, CMD:R,r,w,W,V"
#define VERSIONSTRING "AM29F040 READWRITE (based on MEEPROMMER), CMD:a,w,r,i,e,m,v, version 0.02"

// command line
#define ECHO_ON

//XMODEM stuff
#define POSITION_SOH  0
#define POSITION_PACKET 1
#define POSITION_PACKETSLEFT 2
#define POSITION_PAYLOAD 3
#define POSITION_CRC 131
#define POSITION_CRC2 132

#define ACK 0x06
#define EOT 0x04
#define NAK 0x15


#define LOWADDRESSLATCH  A0
#define HIGHADDRESSLATCH A1
#define CE      A3
#define OE      A4
#define UPPERDATABUSOE A5

#define FLASHA16   10
#define FLASHA17   11
#define FLASHA18   12


//4040 stuff
#define COUNTERMASK 0x0fff

#define DS      A0
#define LATCH   A1
#define CLOCK   A2

#define PORTC_DS      0
#define PORTC_LATCH   1
#define PORTC_CLOCK   2

#define PORTC_CE      3
#define PORTC_OE      4
#define PORTC_WE      5
#define PORTC_UPPERDATABUSOE 5


// eeprom stuff
// define the IO lines for the data - bus
#define D0 2
#define D1 3
#define D2 4
#define D3 5
#define D4 6
#define D5 7
#define D6 8
#define D7 9



//a buffer for bytes to burn
#define BUFFERSIZE 133
byte buffer[BUFFERSIZE];
//command buffer for parsing commands
#define COMMANDSIZE 32
char cmdbuf[COMMANDSIZE];

unsigned long startAddress, endAddress;
unsigned long  lineLength, dataLength;
int loop_cmd = 0; // set to >0 for prompt for command mode
int position; // ptr in xmodem packet
int xmodem_prestart;
int available;
int bytes;  // Make sure this is not defined in 'loop'!
byte crchigh,crclow;
int total;
unsigned long flashaddress;
byte upperbyte;  // This is the state of the data in the 3rd 74HC595. Everytime the address bus is updated, it will use this byte for the D15-D8 lines to the 27C160


//define COMMANDS
#define NOCOMMAND    0
#define VERSION      1
#define SET_ADDRESS  2
#define CLEAR_RAM    3

#define READ_HEX    10
#define READ_BIN    11
#define READ_ITL    12

#define WRITE_HEX   20

#define WRITE_BIN   21
#define WRITE_ITL   22
#define WRITE_XMODEM 23

#define FLASH_IDENTIFY 30
#define FLASH_ERASE 31

#define MD5SUM 40
#define BLANKCHECK 41




/*****************************************************************
 *
 *  CONTROL and DATA functions
 *
 ****************************************************************/

// switch IO lines of databus to INPUT state
#ifdef OLD_IO
void data_bus_input() {
   
  pinMode(D0, INPUT);
  pinMode(D1, INPUT);
  pinMode(D2, INPUT);
  pinMode(D3, INPUT);
  pinMode(D4, INPUT);
  pinMode(D5, INPUT);
  pinMode(D6, INPUT);
  pinMode(D7, INPUT);
}
#endif

#ifdef FAST_IO
void      data_bus_input() {
  DDRD = DDRD & 0x03; // leave bits 0 and 1 as is
  DDRB = DDRB & 0xfc; // leave bits 7 to 2 as is. 
}
#endif

#ifdef OLD_IO
//switch IO lines of databus to OUTPUT state
void data_bus_output() {
  pinMode(D0, OUTPUT);
  pinMode(D1, OUTPUT);
  pinMode(D2, OUTPUT);
  pinMode(D3, OUTPUT);
  pinMode(D4, OUTPUT);
  pinMode(D5, OUTPUT);
  pinMode(D6, OUTPUT);
  pinMode(D7, OUTPUT);
}
#endif

#ifdef FAST_IO
void data_bus_output() {
  DDRD = DDRD | 0xfc; // leave bits 0 and 1 as is
  DDRB = DDRB | 0x03; // leave bits 7 to 2 as is. 
}

#endif

#ifdef OLD_IO
//read a complete byte from the bus
//be sure to set data_bus to input before
byte read_data_bus()
{
  return ((digitalRead(D7) << 7) +
          (digitalRead(D6) << 6) +
          (digitalRead(D5) << 5) +
          (digitalRead(D4) << 4) +
          (digitalRead(D3) << 3) +
          (digitalRead(D2) << 2) +
          (digitalRead(D1) << 1) +
          digitalRead(D0));

}
#endif

#ifdef FAST_IO
byte read_data_bus()
{

  byte b = ((PIND & 0xfc) >> 2) | ((PINB & 0x03) << 6);
  return b;
  
}


#endif


#ifdef OLD_IO
//write a byte to the data bus
//be sure to set data_bus to output before
void write_data_bus(byte data)
{
  //2 bits belong to PORTB and have to be set separtely
  digitalWrite(D6, (data >> 6) & 0x01);
  digitalWrite(D7, (data >> 7) & 0x01);
  //bit 0 to 6 belong to bit 2 to 8 of PORTD
  // meeprommer used to just OR PIND ... which is wrong. You need to mask too
  PORTD = (PIND & 0x03) | ( data << 2 );
  //digitalWrite(D0, (data & 0x01));
  //digitalWrite(D1, (data >> 1) & 0x01);
  //digitalWrite(D2, (data >> 2) & 0x01);
  //digitalWrite(D3, (data >> 3) & 0x01);
  //digitalWrite(D4, (data >> 4) & 0x01);
  //digitalWrite(D5, (data >> 5) & 0x01);
  
  
}
#endif

#ifdef FAST_IO
void write_data_bus(byte data)
{
  PORTD = (PORTD & 0x03) | ( data << 2 );
  PORTB = (PORTB & 0xfc) | (data >> 6);  
}

#endif

#ifdef HC595

/*
 * set_bytemode_address_bus - sets the address bus to a byte address when in BYTE read mode. Its basically
 * a wrapper function to set upperbyte correctly
 */
void set_bytemode_address_bus(unsigned long address)
{
  upperbyte = ((address << 7 ) & 0x80); // in byte read mode b7 of the upper byte is the lowest address bit
  
  set_address_bus(address);
}

/* set_address_bus to a 'address'. 
 *  - The address is effectively a 'byte' address but effectively at a word level. eg $00000 is the first word, $00002 is the 
 *    2nd word. 
 *  - If the chip has BYTE* held LOW, then b7 of the upper databus byte (upperbyte) determines which byte of a word we read. eg
 *    if address is $00000 and b7 of upperbyte is 0, then we read the byte from address $00000 of the chip.
 *    if address is $00000 and b7 of upperbyte is 1, then we read the byte from address $00001.
 *    The address has the lowest bit masked off, so setting address to $00001 actually presents address $00000 to the chip. 
 *  
 *  - If the chip has BYTE* HIGH or at the Vpp level, the chip is in word mode. Again $00000 is the first word, $00002 is the
 *    2nd word.
 *    The value of upperbyte is now used as the value to write to D15-D8 of the chip 
 * 
 */
void set_address_bus(unsigned long address)
{

  //get bits 15 to 8 of address
  byte hi = ((address >>9) & 0xff);
  
  //get bits 7 to 0 of address
  byte low = ((address >> 1) & 0xff);

  //disable latch line
  bitClear(PORTC,PORTC_LATCH);

  //shift out the upper databus byte. We will assume the 3rd 74HC595 OE is high (ie tristate)
  // to read data you need to write 0x00 for the low byte, or 0x80 for the high byte to 'upperbyte'
  // if you are writing data then upperbyte needs to be set to the actual D15-D8 value you want to write
  fastShiftOut(upperbyte);
  //shift out highbyte
  fastShiftOut(hi);
  //shift out lowbyte
  fastShiftOut(low);

  //enable latch and set address
  bitSet(PORTC,PORTC_LATCH);

  // Now the top 3 address bits
  digitalWrite(FLASHA16, (address >>17) & 0x01);  // This is the A16 pin of the EPROM, but is technically A17 in byte mode
  digitalWrite(FLASHA17, (address >>18) & 0x01);
  digitalWrite(FLASHA18, (address >>19) & 0x01);
  

}

//faster shiftOut function then normal IDE function (about 4 times)
void fastShiftOut(byte data) {
  //clear data pin
  bitClear(PORTC,PORTC_DS);
  //Send each bit of the myDataOut byte MSBFIRST
  for (int i=7; i>=0; i--)  {
    bitClear(PORTC,PORTC_CLOCK);
    //--- Turn data on or off based on value of bit
    if ( bitRead(data,i) == 1) {
      bitSet(PORTC,PORTC_DS);
    }
    else {      
      bitClear(PORTC,PORTC_DS);
    }
    //register shifts bits on upstroke of clock pin  
    bitSet(PORTC,PORTC_CLOCK);
    //zero the data pin after shift to prevent bleed through
    bitClear(PORTC,PORTC_DS);
  }
  //stop shifting
  bitClear(PORTC,PORTC_CLOCK);
}



#endif // HC595


//short function to set the OE(output enable line of the eeprom)
// attention, this line is LOW - active
#ifdef OLD_IO
void set_oe (byte state)
{
  digitalWrite(OE, state);
}
#endif

#ifdef FAST_IO
void set_oe (byte state)
{
  if (state==LOW) {
     PORTC &= ~_BV(PORTC_OE);
  } else {
     PORTC |= _BV(PORTC_OE);
  }
}
#endif

#ifdef OLD_IO
//short function to set the CE(chip enable line of the eeprom)
// attention, this line is LOW - active
void set_ce (byte state)
{
  digitalWrite(CE, state);
}
#endif

#ifdef FAST_IO
void set_ce (byte state)
{
   if (state==LOW) {
     PORTC &= ~_BV(PORTC_CE);
  } else {
     PORTC |= _BV(PORTC_CE);
  }
}
#endif

#ifdef OLD_IO
//short function to set the WE(write enable line of the eeprom)
// attention, this line is LOW - active
void set_we (byte state)
{
  digitalWrite(WE, state);
}
#endif

#ifdef FAST_IO
void set_we (byte state)
{
   if (state==LOW) {
     PORTC &= ~_BV(PORTC_WE);
  } else {
     PORTC |= _BV(PORTC_WE);
  }
}
#endif


#ifdef FAST_IO
void set_upperdatabusoe (byte state)
{
   if (state==LOW) {
     PORTC &= ~_BV(PORTC_UPPERDATABUSOE);
  } else {
     PORTC |= _BV(PORTC_UPPERDATABUSOE);
  }
}
#endif



//highlevel function to read a byte from a current address
// This uses the special BYTE mode of the EPROM to read both the high and low byte using only the
// Q7 to Q0 pins of the EPROM.

byte read_next_byte(void)
{
  byte	    data = 0;

  //set databus for reading
  data_bus_input();
  
  //Serial.println("read_next_byte");
  // We have to define upperbyte before we set the address

  set_bytemode_address_bus(flashaddress);
  

  
  
  //first disable output
  set_oe(HIGH);
  // un-tristate the 3rd HC595 so that Q7 of the 3rd HC595 drives Q15 of the EPROM. This will select the low byte or high byte
  set_upperdatabusoe(LOW);
  //enable chip select
  set_ce(LOW);
  //enable output
  set_oe(LOW);
  delayMicroseconds(1);

  data = read_data_bus();

  //disable output
  set_oe(HIGH);
  set_ce(HIGH);
  set_upperdatabusoe(HIGH); // tristate Q15 again

  flashaddress++;

  return data;

}

/*
 * write a 'word' to the EPROM.
 *
 *  normally write 'first' to the high byte (D15 to D8) and 'second' to the low byte (D7 to D0)
 *  but if 'swapthem' is true, write 'first' to the low byte (D7 to D0) and 'second' to the high byte (D15 to D8)

*/
byte eprom_write_next_word(byte first, byte second, boolean swapthem)
{
 
  byte high_d15_to_d8, low_d7_to_d0;
  
  high_d15_to_d8 = first;
  low_d7_to_d0 = second;
  
  if (swapthem) {
     high_d15_to_d8 = second;
     low_d7_to_d0 = first;
  }

  // Make sure we set the upperbyte global BEFORE we set the address bus
  upperbyte = high_d15_to_d8;

  // Set the address (and as a consequence D15 to D8). NB flashaddress is a byte mode address (ie. always even)
  set_address_bus(flashaddress);

  //first disable output
  set_oe(HIGH);
  //set databus for writing
  data_bus_output();

  //disable write
  //set_we(HIGH);
  write_data_bus(low_d7_to_d0);
  set_upperdatabusoe(LOW); // enable D15 to D8 onto the bus
  delayMicroseconds(1); // you are meant to do 2 ... but the function call/exit will consume time
  //enable chip select
  set_ce(LOW);
  //enable output
  //set_oe(LOW);
  delayMicroseconds(46); // 50 us pulse for 27C160


  //set_we(HIGH); // latch data
  set_ce(HIGH);
  delayMicroseconds(1);
  data_bus_input();
  set_upperdatabusoe(HIGH); // tristate D15 to D8 again

  flashaddress+=2; // bump up by two


}



/************************************************
 *
 * COMMAND and PARSING functions
 *
 *************************************************/

//waits for a string submitted via serial connection
//returns only if linebreak is sent or the buffer is filled
void readCommand() {
  //first clear command buffer
  for (int i = 0; i < COMMANDSIZE; i++) cmdbuf[i] = 0;
  //initialize variables
  char c = ' ';
  int idx = 0;
  //now read serial data until linebreak or buffer is full
  while (1) {
    if (idx >= COMMANDSIZE) {
      cmdbuf[idx - 1] = 0;
      break;
    }
    if (Serial.available()) {
      c = Serial.read();
      if ((c == '\n') || (c == '\r')) {
        cmdbuf[idx] = 0;
        
        break;
      }
#ifdef ECHO_ON      
      Serial.write(c); // echo it back
#endif
      cmdbuf[idx++] = c;
    }
  }
}

//parse the given command by separating command character and parameters
//at the moment only 5 commands are supported
byte parseCommand() {
  //set ',' to '\0' terminator (command string has a fixed strucure)
  //first string is the command character
  cmdbuf[1]  = 0;
  //second string is startaddress (5 bytes)
  cmdbuf[7]  = 0;
  //third string is a data length (orig file said endaddress (5 bytes))
  cmdbuf[13] = 0;
  //fourth string is length (2 bytes)
  cmdbuf[16] = 0;
  startAddress = hexFiveWord((cmdbuf + 2));
  dataLength = hexFiveWord((cmdbuf + 8));
  lineLength = hexByte(cmdbuf + 14);
  byte retval = 0;
  switch (cmdbuf[0]) {
    case 'a':
      retval = SET_ADDRESS;
      break;
    case 'c':
      retval = CLEAR_RAM;
      break;
    case 'r':
      retval = READ_HEX;
      break;
    case 'w':
      retval = WRITE_XMODEM;
      break;
    case 't':  // write TWO bytes t xxxxx 0aabb
      retval = WRITE_HEX;
      break;  
    case 'v':
      retval = VERSION;
      break;
    case 'i':
      retval = FLASH_IDENTIFY;
      break;
    case 'e':
      retval = FLASH_ERASE;
      break;
    case 'm':
      retval = MD5SUM;
      break;
    case 'b':
      retval = BLANKCHECK;
      break;
    default:
      retval = NOCOMMAND;
      break;
  }

  return retval;
}

/************************************************************
 * convert a single hex digit (0-9,a-f) to byte
 * @param char c single character (digit)
 * @return byte represented by the digit
 ************************************************************/
byte hexDigit(char c)
{
  if (c >= '0' && c <= '9') {
    return c - '0';
  }
  else if (c >= 'a' && c <= 'f') {
    return c - 'a' + 10;
  }
  else if (c >= 'A' && c <= 'F') {
    return c - 'A' + 10;
  }
  else {
    return 0;   // getting here is bad: it means the character was invalid
  }
}

/************************************************************
 * convert a hex byte (00 - ff) to byte
 * @param c-string with the hex value of the byte
 * @return byte represented by the digits
 ************************************************************/
byte hexByte(char* a)
{
  return ((hexDigit(a[0]) * 16) + hexDigit(a[1]));
}

/************************************************************
 * convert a hex word (0000 - ffff) to unsigned int
 * @param c-string with the hex value of the word
 * @return unsigned int represented by the digits
 ************************************************************/
unsigned int hexWord(char* data) {
  if (data[0]==0) {
     return 0;
  }
  return ((hexDigit(data[0]) * 4096) +
          (hexDigit(data[1]) * 256) +
          (hexDigit(data[2]) * 16) +
          (hexDigit(data[3])));
}

// 5 digit hex number
unsigned long hexFiveWord(char* data) {
  if (data[0]==0) {
     return 0;
  }
  return ((hexDigit(data[0])*65536L) +
          (hexDigit(data[1]) * 4096L) +
          (hexDigit(data[2]) * 256L) +
          (hexDigit(data[3]) * 16L) +
          (hexDigit(data[4])));
}
/************************************************
 *
 * INPUT / OUTPUT Functions
 *
 *************************************************/


/**
 * read a data block from eeprom and write out a hex dump
 * of the data to serial connection
 * @param from       start address to read fromm
 * @param to         last address to read from
 * @param linelength how many hex values are written in one line
 **/
void read_block(unsigned long  from, unsigned long  to, int linelength)
{
  flashaddress=from;
  set_bytemode_address_bus(flashaddress); // technically you dont need to do this as read_next_byte does it anyway

  //count the number fo values that are already printed out on the
  //current line
  int		    outcount = 0;
  //loop from "from address" to "to address" (included)
  for (unsigned long address = from; address <= to; address++) {
    if (outcount == 0) {
      //print out the address at the beginning of the line
      //Serial.println();
      Serial.print("\r\n");
      Serial.print("0x");
      printAddress(address);
      Serial.print(" : ");
    }
    //print data, separated by a space
    byte data = read_next_byte();
    printByte(data);
    Serial.print(" ");
    outcount = (++outcount % linelength);

  }
  //print a newline after the last data line
  Serial.println();

}

void read_md5(unsigned long startAddress, unsigned long dataLength) {
  unsigned long  bytesleft;
  int bytestoprocess;
  int i;
  char md5buffer[64];
  char md5str[33];
  MD5_CTX context;
  unsigned char digest[16];
  
  flashaddress=startAddress;
  set_bytemode_address_bus(flashaddress);

  md5str[0] = '\0';
  MD5Init(&context);
  
  while (flashaddress < (startAddress + dataLength) ) {
     // Work out how many bytes are left
     bytesleft = ((startAddress + dataLength) - flashaddress);
     bytestoprocess=64;
     if (bytesleft < 64) {
        bytestoprocess = bytesleft;
     }
     for (i=0;i<bytestoprocess;i++) {
        md5buffer[i] = (char) read_next_byte();
     }
     MD5Update(&context, md5buffer, bytestoprocess);
     // No need to bump up flash address as read_next_byte does it.
  }

  MD5Final(digest, &context);

  make_digest(md5str, digest, 16);
  Serial.println(md5str);
  Serial.println();
  
}

void blank_check(unsigned long startAddress, unsigned long dataLength) {
  byte b;
  boolean found_non_ff = false;
    
  flashaddress=startAddress;
  set_bytemode_address_bus(flashaddress);
  
  while (flashaddress < (startAddress + dataLength) ) {
     b = read_next_byte();
     if (b != 0xff) {
         found_non_ff = true;
         flashaddress--;
         break;
     }
  }   
  if (found_non_ff) {
    Serial.print(F("Found byte "));
    printByte(b);
    Serial.print(F(" at "));
    printAddress(flashaddress);
    Serial.println();
  } else {
    Serial.println(F("Scan complete. Only FFs found"));
  }
  
}

// debug function
void write_two_bytes(unsigned long from, byte a, byte b)
{
  // prep
  //special_flash_write(0x0000, 0xf0); // read-reset
  
  flashaddress=from;
  
  eprom_write_next_word(a,b,LITTLEENDIAN);
  
  //print a newline after the last data line
  Serial.println();
  Serial.println("Put Vcc back to 5V, and BYTE/Vpp back to GND");
}



/**
 * print out a 20 bit word as 5 character hex value
 **/
void printAddress(unsigned long address) {
  if (address < 0x0010) Serial.print("0");
  if (address < 0x0100) Serial.print("0");
  if (address < 0x1000) Serial.print("0");
  if (address < 0x10000) Serial.print("0");
  Serial.print(address, HEX);

}

/**
 * print out a byte as 2 character hex value
 **/
void printByte(byte data) {
  if (data < 0x10) Serial.print("0");
  Serial.print(data, HEX);
}



/**
 *  Calc the Xmodem CRC16
 */
unsigned int calccrc(byte *ptr, int count)
{
    unsigned int  crc;
    int i;
    
    crc = 0;
    
    while (--count >= 0)
    {
        crc = crc ^ (unsigned int) (*ptr++ << 8);
        i = 8;
        do
        {
            if (crc & 0x8000)
                crc = crc << 1 ^ 0x1021;
            else
                crc = crc << 1;
        } while(--i);
    }
    return (crc);
}



/************************************************
 *
 * MAIN
 *
 *************************************************/
void setup() {
  data_bus_input();
  flashaddress=0;  // master address for sequential data
  upperbyte=0;  // b7 is the upper/lower togger when we are reading. b7=0 means low byte of word. b7=1 means high byte.

  //Turn LED OFF
  //pinMode(LEDPIN, OUTPUT);
  //digitalWrite(LEDPIN, LOW);

#ifdef HC595
  pinMode(DS,OUTPUT);
  pinMode(LATCH,OUTPUT);
  pinMode(CLOCK,OUTPUT);
#endif
  
  pinMode(CE, OUTPUT);
  digitalWrite(CE, HIGH);

  pinMode(OE, OUTPUT);
  digitalWrite(OE, HIGH);

  pinMode(UPPERDATABUSOE, OUTPUT);
  digitalWrite(UPPERDATABUSOE,HIGH);  // HIGH means tristate the 3rd 74HC595
  
  pinMode(FLASHA16, OUTPUT);
  pinMode(FLASHA17, OUTPUT);
  pinMode(FLASHA18, OUTPUT);


  Serial.begin(115200);
  Serial.println(F("Started. Note: all hex addresses are five digits"));
  Serial.println(F("  r nnnnn mmmmm - show mmmmm bytes at address nnnnn"));
  Serial.println(F("  w nnnnn - write to eprom using xmodem transfer"));
  //Serial.println("  e - erase flash rom");
  //Serial.println("  i - identify flash rom");
  Serial.println(F("  m nnnnn mmmmm - md5sum rom content starting at nnnnn for mmmmmm bytes long"));
  Serial.println(F("  b nnnnn mmmmm - Check for FFs from nnnnn for mmmmmm bytes long"));
  Serial.println();
  if (LITTLEENDIAN) {
    Serial.println(F("WARNING: All functions that WRITE to the EPROM will be in little endian mode"));
    Serial.println(F("This means the first byte from your file goes to the low byte (D7 to D0) and"));
    Serial.println(F("and the second byte from your file goes to the high byte (D15 to D8)"));
  } else {
    Serial.println(F("WARNING: All functions that WRITE to the EPROM will transfer in big endian mode"));
    Serial.println(F("This means the first byte from your file goes to the high byte (D15 to D8) and"));
    Serial.println(F("and the second byte from your file goes to the low byte (D7 to D0)"));
    
  }
  
  
  
}







/**
 * main loop, that runs invinite times, parsing a given command and
 * executing the given read or write requestes.
 **/
void loop() {
  byte a, b;

  byte xmodem_byte;
  unsigned int c;

  unsigned int crc;

  if (loop_cmd == NOCOMMAND) {
    Serial.print("\r% "); // show the prompt
    readCommand();
#ifdef ECHO_ON    
    Serial.println(""); // only required in ECHO mode
#endif    
    byte cmd = parseCommand();
    
    switch (cmd) {
      // clear ram is a leftover from when I used some of this code to write to a static ram.

      // setting the address like this is mainly used for testing to see if your 74HC595s or 74LS374s are working.
      case SET_ADDRESS:
        // Set the address bus to an arbitrary value.
        // Useful for debugging shift-register wiring, byte-order.
        // e.g. a,00FF
        Serial.print("Setting address bus to 0x");
        Serial.println(cmdbuf + 2);
        Serial.println(startAddress);
        set_address_bus(startAddress);
        set_upperdatabusoe(LOW);
        
        

        break;
         
      case READ_HEX:
        //set a default if needed to prevent infinite loop
        if (lineLength == 0) lineLength = 16;
        if (dataLength == 0) {
           dataLength = 0x80;
        }
        endAddress = startAddress + dataLength - 1;
        read_block(startAddress, endAddress, lineLength);
        break;
      case MD5SUM:
        //set a default if needed to prevent infinite loop
        
        if (dataLength == 0) {
           dataLength = 0x80;
        }
        endAddress = startAddress + dataLength - 1;
        Serial.println("Reading md5");
        read_md5(startAddress, dataLength);
        break;
     case BLANKCHECK:
        //set a default if needed to prevent infinite loop
        
        if (dataLength == 0) {
           dataLength = 0x80;
        }
        endAddress = startAddress + dataLength - 1;
        Serial.println("Checking for anything that isnt an FF");
        blank_check(startAddress, dataLength);
        break;
 
      case READ_ITL:
        break;

      case WRITE_HEX: // t xxxxx 0yyzz  Write at xxxxx the two bytes yy and zz
        // simple test write of two bytes at the address
        a = ((dataLength >> 8) & 0xff);
        b = dataLength & 0xff;
        write_two_bytes(startAddress, a, b);
        break;
      case WRITE_XMODEM:
        Serial.println(F("Put the Vcc to 6.25V, then put Vpp to 12.5V"));
        Serial.println(F("Pause 15 secs to give you some time"));
        delay(15000);
        Serial.println(F("Please start an xmodem transfer in your terminal ... "));
        Serial.println(F("When the transfer finishes change BYTE/Vpp back to GND, and switch Vcc back to 5V."));
        position = 0;
        bytes = 0;
        total=0;
        flashaddress=startAddress;
        //special_flash_write(0x0000, 0xf0); // read-reset
        set_address_bus(startAddress);
        xmodem_prestart = 1;

        loop_cmd = WRITE_XMODEM;
        break;
      case VERSION:
        Serial.println(VERSIONSTRING);
        break;
      default:
        break;
    }
    
    
  } else if (loop_cmd == WRITE_XMODEM) {
    if (xmodem_prestart) {  // In prestart mode we are just waiting for the sending device to start sending.
      Serial.write('C');
      delay(2000);
      if (Serial.available()) {
        xmodem_prestart = 0;
      }
    } else {
      //should be a byte read to read.
      while (Serial.available()) {
         xmodem_byte = Serial.read();
         switch(position) {
            case POSITION_SOH:
               bytes=0;
               if (xmodem_byte == EOT) {
                  Serial.write(ACK);  // send ACK
                  loop_cmd = NOCOMMAND;
                  //Serial.print("Wrote ");
                  //Serial.println(total,DEC);
               } 
               break;
            case POSITION_CRC:
               crchigh = xmodem_byte;
               break;
            case POSITION_CRC2:
               crc = calccrc(buffer,128);
               if ( (((crc >> 8) &0xff) == crchigh) && ( (crc & 0xff) == xmodem_byte) ) {
                  for (int i=0;i<128;i+=2) {
                     eprom_write_next_word(buffer[i],buffer[i+1],LITTLEENDIAN); //true because we want to byte swap
                     total+=2;
                  }
                  Serial.write(ACK);  // send ACK
               } else {
                  Serial.write(NAK); // CRC failed. Send NAK, and because we never wrote anything to SRAM, the address counter wont have moved
               }
               position = -1;
               break;
            default:
               if ((position >=POSITION_PAYLOAD) && (position <POSITION_CRC)) {
                  buffer[bytes++] = xmodem_byte;
               }
               break;
          
         }
         position++;
      }  
    }
  }
}





